import sys
from youtrack.connection import Connection
from fbugz.fbDbClient import FBClient
from youtrack import User, Subsystem, Version
import fbugz

def main():
    try:
        target_url, target_login, target_pass, fb_db, fb_host, fb_port, fb_login, fb_pass = sys.argv[1:9]
        project_names = sys.argv[9:]
    except:
        print 'Usage : fb2youtrackMySql.py target_url target_login target_pass fb_db fb_host fb_port fb_login fb_pass project_names'
        exit()
    fb2youtrack(target_url, target_login, target_pass, fb_db, fb_host, fb_port, fb_login, fb_pass, project_names)

def to_yt_user(fb_user):
    yt_user = User()
    yt_user.login = fb_user.login
    yt_user.email = fb_user.email
    yt_user.fullName = fb_user.login
    return yt_user

def to_yt_subsystem(area) :
    yt_subsystem = Subsystem()
    yt_subsystem.name = area.name
    yt_subsystem.isDefault = False
    yt_subsystem.defaultAssignee = area.person_owner
    return yt_subsystem

def to_yt_version(milestone) :
    yt_version = Version()
    yt_version.name = milestone.name
    yt_version.isReleased = milestone.inactive
    yt_version.isArchived = milestone.deleted
    yt_version.releaseDate = milestone.releaseDate
    return yt_version

def fb2youtrack(target_url, target_login, target_pass, fb_db, fb_host, fb_port, fb_login, fb_pass, project_names):
    target = Connection(target_url, target_login, target_pass)
    source = FBClient(fb_host, fb_port, fb_login, fb_pass, fb_db)

    fb_users = source.get_users()
    yt_users = list([])
    max_count = 100
    print 'Importing users'
    for fb_user in fb_users:
        print 'Processing user [ %s ]' % fb_user.login
        yt_users.append(to_yt_user(fb_user))
        if (len(yt_users) >= max_count):
            target.importUsers(yt_users)
            yt_users = list([])
    target.importUsers(yt_users)
    print 'Importing users finished'

    print 'Importing custom fields'
    fb_custom_fields = source.get_custom_fields()
    for custom_field in fb_custom_fields:
        print 'Processing custom field [ %s ]' % custom_field.name
        cf_type = fbugz.CF_TYPES[custom_field.type]
        if (cf_type in ['emum[1]', 'enum[*]']):
            bundle_name = custom_field.name + '_bundle'
            print 'Creating bundle with name [ %s ] for custom field [ %s ]' % (bundle_name, custom_field.name)
            try:
                target.createEnumBundleDetailed(bundle_name, custom_field.possible_values)
            except:
                print "Can't create bundle with name [ %s ] (maby because it already exists)" % bundle_name
        try:
            target.createCustomFieldDetailed(custom_field.name, cf_type, False, True)
        except:
            print "Can't create custom field with name [ %s ] (maby because it already exists)" % custom_field.name
    print 'Importing custom fields finished'

    for project_name in project_names :
        print 'Processing project [ %s ]' % project_name
        project_id = source.get_project_id_by_name(project_name)
        try :
            target.getProject(str(project_id))
        except :
            target.createProjectDetailed(project_id, project_name, "", target_login)

        print 'Importing areas to project [ %s ]' % project_name
        areas = source.get_areas(project_id)
        for area in areas :
            print 'Processing subsystem [ %s ]' % area.name
            try :
                target.createSubsystem(str(project_id), to_yt_subsystem(area))
            except :
                print "Can't create subsystem with name [ %s ] in project [ %s ] (maby because it already exists)" % (area.name, project_name)
        print 'Importing areas to project [ %s ] finished' % project_name

        print 'Importing versions to peroject [ %s ]' % project_name
        milestones = source.get_milestones(project_id)
        for milestone in milestones :
            print 'Processing version [ %s ]' % milestone.name
            try :
                target.createVersion(str(project_id), to_yt_version(milestone))
            except :
                print "Can't create version with name [ %s ] in project [ %s ] (maby because it already exists)" % (milestone.name, project_name)
        print 'Importing versions finished'
            
        

if __name__ == '__main__':
    main()