from fbugz import FBArea, FBUser, FBMilestone, FBCustomField
import MySQLdb
import MySQLdb.cursors

area_types = dict({
            0   :   "Normal",
            1   :   "Not Spam",
            2   :   "Undecided",
            3   :   "Spam"
                          })


class FBClient:
    def __init__(self, host, port, login, password, db_name):
       self._sql_cnx = MySQLdb.connect(host = host, port = port, user = login, passwd = password,
                                       db = db_name, cursorclass = MySQLdb.cursors.DictCursor)

    def get_project_id_by_name(self, project_name) :
        project_cursor = self._sql_cnx.cursor()
        id_row = 'ixProject'
        request = 'SELECT %s FROM Project WHERE sProject = %s' % (id_row, project_name)
        project_cursor.execute(request)
        return project_cursor.fetchone()[id_row]

    def get_users(self) :
        user_cursor = self._sql_cnx.cursor()
        full_name_row = 'sFullName'
        email_row = 'sEmail'
        request = 'SELECT %s, %s FROM Person'
        user_cursor.execute(request)
        users = list([])
        for row in user_cursor :
            user = FBUser(row[full_name_row])
            user.email = row[email_row]
            users.append(user)
        return users

    def get_areas(self, ix_project) :
        area_cursor = self._sql_cnx.cursor()
        owner_row = 'ixPersonOwner'
        area_name_row = 'sArea'
        type_row = 'nType'
        request = 'SELECT %s, %s, %s FROM Area WHERE (ixProject = %d)' % (owner_row, area_name_row, type_row, ix_project)
        area_cursor.execute(request)
        areas_to_return = list([])
        for row in area_cursor :
            area = FBArea(row[area_name_row])
            area.n_type = area_types[row[type_row]]
            area.person_owner = self._get_person_name_by_id(row[owner_row])
            areas_to_return.append(area)
        return areas_to_return

    def get_milestones(self, ix_project) :
        milestone_cursor = self._sql_cnx.cursor()
        name_row = 'sFixFor'
        release_date_row = 'dt'
        deleted_row = 'fDeleted'
        inactive_row = 'finactive'
        request =  'SELECT %s, %s, %s, %s FROM FixFor ' % (name_row, release_date_row, deleted_row, inactive_row)
        request += 'WHERE (ixProject = %d) OR (ixProject = -1)' % ix_project
        milestone_cursor.execute(request)
        milestones_to_return = list([])
        for row in milestone_cursor :
            milestone = FBMilestone(row[name_row])
            milestone.deleted = (row[deleted_row] == 1)
            milestone.inactive = (row[inactive_row] == 1)
            milestone.release_date = self._get_date(row[release_date_row])
            milestones_to_return.append(milestone)
        return milestones_to_return

    def get_custom_fields(self) :
        cf_cursor = self._sql_cnx.cursot()
        name_row = 'sName'
        column_name_row = 'sColumnName'
        type_row = 'nType'
        values_row = 'srgsPoosibleValues'
        request = 'SELECT %s, %s, %s, %s FROM Plugin_18_CustomField' % (name_row, column_name_row, type_row, values_row)
        cf_cursor.execute(request)
        custom_fields = list([])
        for row in cf_cursor :
            cf = FBCustomField(row[name_row], row[column_name_row])
            cf.type = row[type_row]
            if (cf.type == 1) :
                for v in row[values_row].split('\n') :
                    cf.possible_values.append(v.strip())
            elif (cf.type == 5) :
                cf.possible_values.append('True')
                cf.possible_values.append('False')
            custom_fields.append(cf)
        return custom_fields


    def _get_person_name_by_id(self, id) :
        person_cursor = self._sql_cnx.cursor()
        name_row = 'sFullName'
        request = 'SELECT %s FROM Person WHERE ixPerson = %d' % (name_row, id)
        person_cursor.execute(request)
        row = person_cursor.fetchone()
        if (row is not None) :
            return row[name_row]
        else :
            return None


    def _get_date(self, date_time) :
        pass