from fbugz.fogbugz import FogBugz
from fbugz import FBUser, FBArea, FBMilestone, FBIssue, FBComment
import time
from fbugz import defaultFBugz
import fbugz
from datetime import datetime
import calendar

class FBClient(object):
    def __init__(self, source_url, source_login, source_password):
        self._source_url = source_url
        self._client = FogBugz(source_url)
        self._client.logon(source_login, source_password)
        self._case_ids = []

    def list_project_names(self):
        projects = dict([])
        for p in self._client.listProjects().findAll('project'):
            projects[p.sproject.string.strip()] = p.ixproject.string.strip()
        return projects

    def get_users(self):
        self._users = []
        for p in self._client.listPeople(fIncludeNormal=1).findAll('person'):
            self._users.append(self._create_user(p, 'Normal'))

        for p in self._client.listPeople(fIncludeNormal=0, fIncludeDeleted=1).findAll('person'):
            self._users.append(self._create_user(p, 'Deleted'))

        for p in self._client.listPeople(fIncludeNormal=0, fIncludeVirtual=1).findAll('person'):
            self._users.append(self._create_user(p, 'Virtual'))

        for p in self._client.listPeople(fIncludeNormal=0, fIncludeCommunity=1).findAll('person'):
            self._users.append(self._create_user(p, 'Community'))

        return self._users

    def get_areas(self, ix_project):
        result = []
        for a in self._client.listAreas(ixproject=ix_project).findAll('area'):
            area = FBArea(a.sarea.string.encode('utf-8').decode('utf-8'))
            owner = a.spersonowner.string
            if (owner is None):
                area.person_owner = None
            else:
                area.person_owner = owner.encode('utf-8').decode('utf-8')
            result.append(area)
        return result

    def get_milestones(self, ix_project):
        result = []
        for m in self._client.listFixFors(ixProject=ix_project, fIncludeDeleted=1, fIncludeReallyDeleted=1).findAll(
                "fixfor"):
            milestone = FBMilestone(m.sfixfor.string.encode('utf-8').decode('utf-8'))
            dt = m.dt.string
            milestone.release_date = self._to_unix_date(dt)
            milestone.inactive = bool(m.fdeleted.string)
            result.append(milestone)
        return result

    def get_issues(self, project_name, start_id, num):
        result = []
        cols_string = "ixBug,sArea,sPersonAssignedTo,ixBugParent,sCategory,dtClosed,sComputer,sVersion,dtDue,sFixFor,"
        cols_string += "sLatestTextSummary,fOpen,dtOpened,dtResolved,ixPriority,sStatus,sOriginalTitle,sTitle,tags,"
        cols_string += "ixPersonOpenedBy,ixPersonResolvedBy,hrsCurrEst"

        for i in self._client.search(cols=cols_string, q=(
        'case:%d..%d project:"%s"' % (start_id, start_id + num - 1, project_name))).findAll('case'):
            ix_bug = i.ixbug.string
            self._case_ids.append(ix_bug)
            issue = FBIssue(ix_bug)
            issue.area = i.sarea.string.encode('utf-8')
            assignee = i.spersonassignedto.string
            if assignee is not None :
                if assignee == 'CLOSED':
                    issue.assignee = self._find_user_login(i.ixpersonresolvedby.string)
                else:
                    issue.assignee = assignee.encode('utf-8')
            parent_id = i.ixbugparent.string
            if parent_id != '0':
                issue.bug_parent = parent_id
            issue.category = i.scategory.string
            issue.estimate = i.hrscurrest.string
            issue.closed = self._to_unix_date(i.dtclosed.string)
            computer = i.scomputer.string
            if computer is not None:
                issue.computer = computer.encode('utf-8')
            sversion_string = i.sversion.string
            if sversion_string is not None:
                issue.version = sversion_string.encode('utf-8')
            due_dt = i.dtdue.string
            if (due_dt is None) or (due_dt.strip() == ''):
                issue.due = None
            else:
                issue.due = self._to_unix_date(due_dt)
            issue.fix_for = i.sfixfor.string.encode('utf-8')
            latest_summary = i.slatesttextsummary.string
            if latest_summary is not None :
                issue.latest_text_summary = latest_summary.encode('utf-8')
            issue.open = bool(i.fopen.string)
            issue.opened = self._to_unix_date(i.dtopened.string)
            issue.resolved = self._to_unix_date(i.dtresolved.string)
            issue.priority = int(i.ixpriority.string)
            issue.status = i.sstatus.string
            original = i.soriginaltitle.string
            if original is not None:
                issue.original_title = original.encode('utf-8')
            else:
                issue.original_title = None
            issue.title = i.stitle.string.encode('utf-8')
            events = self._client.search(q=ix_bug, cols='events').findAll('event')
            issue.attachments = self._get_attachments_from_events(events)
            issue.comments = self._get_comments_from_events(events)
            for tag in i.tags.findAll('tag'):
                issue.tags.append(tag.string)
            issue.reporter = self._find_user_login(i.ixpersonopenedby.string)
            result.append(issue)

        return result

    def get_issue_project_id(self, ix_bug):
        return self._client.search(q=ix_bug, cols='ixProject').case.ixproject.string

    def _get_comments_from_events(self, events):
        comments = []
        for event in events:
            if hasattr(event, 's'):
                comment = FBComment()
                content = event.s.string
                if (content is None) or (content.strip() == ''):
                    continue
                comment.text = content.encode('utf-8')
                comment.author = event.sperson.string.encode('utf-8')
                comment.date = self._to_unix_date(event.dt.string)
                comments.append(comment)
        return comments

    def _get_attachments_from_events(self, events):
        attachments = []
        for event in events:
            for a in event.findAll('attachment'):
                attach = fbugz.FBAttachment(self._source_url, a.surl.string)
                attach.authorLogin = event.sperson.string
                attach.token = self._client.get_token()
                attachments.append(attach)
        return attachments


    def _create_user(self, p, type):
        person = FBUser(p.sfullname.string)
        if (len(p.semail.string)):
            person.email = p.semail.string
        person.user_type = type
        person.id = p.ixperson.string
        return person

    def _find_user_login(self, id):
        if (self._users is None):
            self.get_users()
        for user in self._users:
            if (user.id.string == id):
                return user.login.string.encode('utf-8')
        return None

    def _to_unix_date(self, date_string):
        if (date_string is None):
            dt = datetime.now()
        else:
            dt = datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%SZ')
        return str(calendar.timegm(dt.timetuple()) * 1000)